@font-face {
    font-family: 'Sage';
    src: url('../fonts/sage-bold.woff2') format('woff');
}

@font-face {
    font-family: 'Sherpa';
    src: url('../fonts/sherpa.woff2') format('woff');
}

/* h1 {
    font-family: 'Sage', sans-serif !important;
    font-size: 44px !important;
    color: #000 !important;
}

:root {
    --md-text-font: "Sherpa";
} */

.md-header__button.md-logo img,
.md-header__button.md-logo svg {
    height: 3rem !important;
    margin-bottom: 0em;
    padding-bottom: 0;
}

.no-bottom-margin {
    margin-bottom: 0 !important;
}

.maiusc {
    text-transform: uppercase;
}

.center {
    display: block;
    margin-left: auto;
    margin-right: auto;
}

.doc-attribute {
    border-top: 1px solid #ccc;
}

.doc-method {
    border-top: 1px solid #ccc;
}

.doc-function {
    border-top: 1px solid #ccc;
}

.doc-class {
    border-top: 5px solid var(--md-code-bg-color);
}

/* Remove the `In` and `Out` block in rendered Jupyter notebooks */
.md-container .jp-Cell-outputWrapper .jp-OutputPrompt.jp-OutputArea-prompt,
.md-container .jp-Cell-inputWrapper .jp-InputPrompt.jp-InputArea-prompt {
    display: none !important;
}

/* Fix /page#foo going to the top of the viewport and being hidden by the navbar */
html {
    scroll-padding-top: 50px;
}

/* Emphasise sections of nav on left hand side */
nav.md-nav {
    padding-left: 5px;
}

nav.md-nav--secondary {
    border-left: revert !important;
}

.md-nav__title {
    font-size: 0.9rem;
}

.md-nav__item--section>.md-nav__link {
    font-size: 0.9rem;
}

/* More space at the bottom of the page */

.md-main__inner {
    margin-bottom: 1.5rem;
}


/* Change font sizes */
html {
    /* Decrease font size for overall webpage
       Down from 137.5% which is the Material default */
    font-size: 110%;
}

.md-typeset .admonition {
    /* Increase font size in admonitions */
    font-size: 100% !important;
}

.md-typeset details {
    /* Increase font size in details */
    font-size: 100% !important;
}

.md-typeset h1 {
    font-size: 1.6rem;
}

.md-typeset h2 {
    font-size: 1.5rem;
}

.md-typeset h3 {
    font-size: 1.3rem;
}

.md-typeset h4 {
    font-size: 1.1rem;
}

.md-typeset h5 {
    font-size: 0.9rem;
}

.md-typeset h6 {
    font-size: 0.8rem;
}


/* Highlight functions, classes etc. type signatures. Really helps to make clear where
   one item ends and another begins. */

[data-md-color-scheme="default"] {
    --doc-heading-color: #DDD;
    --doc-heading-border-color: #CCC;
    --doc-heading-color-alt: #F0F0F0;
}

[data-md-color-scheme="slate"] {
    --doc-heading-color: rgb(25, 25, 33);
    --doc-heading-border-color: rgb(25, 25, 33);
    --doc-heading-color-alt: rgb(33, 33, 44);
    --md-code-bg-color: rgb(38, 38, 50);
}

h4.doc-heading {
    /* NOT var(--md-code-bg-color) as that's not visually distinct from other code blocks.*/
    background-color: var(--doc-heading-color);
    border: solid var(--doc-heading-border-color);
    border-width: 1.5pt;
    border-radius: 2pt;
    padding: 0pt 5pt 2pt 5pt;
}

h5.doc-heading,
h6.heading {
    background-color: var(--doc-heading-color-alt);
    border-radius: 2pt;
    padding: 0pt 5pt 2pt 5pt;
}